from os.path import join

import mekpie.messages as messages

from .definitions import Config
from .util        import panic, tab, check_is_file, check_is_dir, type_name, exec_file
from .structure   import get_main_path, get_mekpy_path

def get_config(options):
    return config_from_dict(exec_file(get_mekpy_path(), { 'options': options }), options)

def config_from_dict(config_dict, options):
    return check_config(Config(
        name    = config_dict.get('name',  None),
        main    = config_dict.get('main',  None),
        libs    = config_dict.get('libs',  []),
        cc      = config_dict.get('cc',    'gcc/clang'),
        cmd     = config_dict.get('cmd',   'cc'),
        dbg     = config_dict.get('dbg',   'gdb'),
        flags   = config_dict.get('flags', []),
        options = options,
    ))

def check_config(config):
    check_name(config.name)
    check_main(config)
    check_libs(config.libs)
    check_cc(config.cc)
    check_cmd(config.cc)
    check_dbg(config.dbg)
    check_flags(config.flags)
    return config

def check_name(name):
    check_type('name', name, str)

def check_main(config):
    check_type('main', config.main, str)
    check_is_file(get_main_path(config.main))

def check_libs(libs):
    check_type('libs', libs, list)
    for lib in libs:
        check_type('libs', lib, str)

def check_cc(cc):
    check_type('cc', cc, str)

def check_cmd(cmd):
    check_type('cmd', cmd, str)

def check_dbg(dbg):
    check_type('dbg', dbg, str)

def check_flags(flags):
    check_type('flags', flags, list)
    for flag in flags:
        check_type('flags', flag, str)

def check_type(name, value, *expected_types):
    if all([type(value) != exp for exp in expected_types]):
        panic(messages.type_error.format(
            name,
            ' or '.join([exp.__name__ for exp in expected_types]),
            type_name(value),
            tab(get_description(name))
        ))

def get_description(name):
    return {
        'name'                   : '`name` specifies the name of the project',
        'main'                   : '`main` specifies the .c file continaing `main`',
        'libs'                   : '`libs` speicfies and libraries to load',
        'cc'                     : '`cc` specifies the c compiler configuration to use',
        'cmd'                    : '`cmd` specified the c compiler command',
        'dbg'                    : '`dbg` specifies the debugger to use',
        'flags'                  : '`flags` specifies additional compiler flags',
    }[name]


from re      import sub
from os      import walk, mkdir, rename, remove
from sys     import stderr
from shutil  import rmtree
from os.path import isfile, isdir, join, basename, splitext, exists
from filecmp import dircmp

import mekpie.debug    as debug
import mekpie.messages as messages

# Debug
# ---------------------------------------------------------------------------- #

def panic(message=None):
    if message is None:
        exit(1)
    errprint(f'\n{message.strip()}\n\n')
    if debug.debug:
        raise Exception('Debug')
    else:
        exit(1)
    

def log(message):
    errprint(f' -- {tab(str(message)).strip()}\n') if debug.debug else None
    return message

def errprint(string):
    stderr.write(string)
    stderr.flush()

# Collections
# ---------------------------------------------------------------------------- #

def empty(collection):
    return len(collection) == 0

def car(collection):
    if not empty(collection):
        return collection[0]

def last(collection):
    if not empty(collection):
        return collection[-1]

def cdr(collection):
    if not empty(collection):
        return collection[1:]

def cons(item, collection):
    return [item] + collection

def shift(collection, n=1):
    for _ in range(n):
        if not empty(collection):
            collection.pop(0)

def flatten(collection):
    return sum(collection, [])

def split(collection, item):
    if item not in collection:
        return collection, []
    index  = collection.index(item)
    first  = collection[:index]
    second = collection[index + 1:] 
    return first, second


# Strings
# ---------------------------------------------------------------------------- #

def tab(string, spaces=4):
    return sub(r'^|\n', '\n' + (spaces * ' '), string)

def underline(element, collection):
    top    = ' '.join(collection)
    bottom = ' '.join(underlined_collection(element, collection))
    return f'{top}\n{bottom}'

def underlined_collection(underlined_element, collection):
    def underline_or_hide(element):
        rep = '^' if element == underlined_element else ' '
        return sub(r'.', rep, str(element))
    return map(underline_or_hide, collection)

# Files
# ---------------------------------------------------------------------------- #

def smkdir(path):
    if not exists(path):
        mkdir(path)

def srmtree(path):
    if exists(path):
        rmtree(path)


def smv(source, destination):
    remove(destination)
    if exists(source):
        rename(source, destination)

def list_files(path, with_filter=None, with_ext=None, recursive=False):
    if with_filter is None:
        with_filter = lambda _ : True
    if with_ext is not None:
        with_filter = lambda filename : filename.endswith(with_ext)
    return list(filter(with_filter, list_all_files(path)))

def list_all_files(path):
    return flatten([[join(pre, post)
            for post
            in posts]
        for (pre, _, posts)
        in walk(path)
    ])

def list_all_dirs(path):
    return flatten([[join(path, pre)
            for pre 
            in pres]
        for (_, pres, posts)
        in walk(path)
    ])

def filename(path):
    return splitext(basename(path))[0]

def file_as_str(path):
    log(f'Reading the contents of {path}...')
    with open(path) as resource:
        return resource.read()

def remove_contents(path):
    log(f'Deleting the contents of {path}...')
    srmtree(path)
    smkdir(path)

def check_is_file(path):
    if isfile(path):
        return path
    panic(messages.file_not_found.format(path))

def check_is_dir(path):
    if isdir(path):
        return path
    panic(messages.directory_not_found.format(path))

def same_dir(dir1, dir2):
    def recursive(dcmp):
        if dcmp.diff_files:
            return False
        return all([recursive(sub_dcmp)
            for sub_dcmp
            in dcmp.subdirs.values()
        ])
    return recursive(dircmp(dir1, dir2))

def exec_str(source, handle, ctx={}):
    try:
        exec(source, ctx)
    except Exception as err:
        panic(messages.execution_error.format(handle, tab(str(err))))
    return ctx

def exec_file(path, ctx={}):
    return exec_str(file_as_str(path), path, ctx)

# Types
# ---------------------------------------------------------------------------- #

def type_name(x):
    return type(x).__name__



from sys        import stdout, stderr
from subprocess import run, PIPE

import mekpie.messages as messages

from .util import panic, log

def lrun(args, quiet=False, error=True):
    log('Running...\n' + serialize_command(args))
    try:
        if quiet:
            if run(args, stdout=PIPE, stderr=PIPE).returncode != 0:
                raise OSError
        else:
            if run(args).returncode != 0:
                raise OSError
        return True
    except KeyboardInterrupt:
        exit(1)
    except OSError:
        if error:
            panic(messages.failed_program_call.format(serialize_command(args)))
        else:
            False

def serialize_command(args):
    return '$ ' + ' '.join(args)



from .definitions import VERSION

usage = '''
Usage:
    mekpie [options] <command> [<args>...]

Options:
    -h, --help      Display this message
    -V, --version   Print version info and exit
    -v, --verbose   Use verbose output
    -q, --quiet     No output to stdout
    -r, --release   Run the command for release
    -d, --developer Run mekpie in developer mode
    -c, --changedir Run the mekpie command in the provided directory

Commands:
    new  <name>     Create a new mekpie project
    test <names...> Build and execute tests or test
    clean           Remove artefacts from previous builds
    build           Compile the current project
    run             Build and execute main
    debug           Build and execute under a debugger
    dist            Create an executable in the project root

'''

version = f'''
    mekpie version {VERSION}
'''

no_command = '''
Invalid command: No command provided!

    mekpie expects a command following its invocation.
    The following commands are available:

        help        Display a help message
        version     Print version info and exit
        new <name>  Create a new mekpie project
        init <name> Create a project in an existing directory
        test <name> Build and execute tests or test
        clean       Remove artefacts from previous builds
        build       Compile the current project
        run         Build and execute main
        debug       Build and execute under a debugger
        dist        Create an executable in the project root
'''

repeated_option = 'Repeated option: option {} appears twice!'

too_many_arguments = 'Unexpected argument: Too many arguments provided!'

unknown_argument = 'Unexpected argument: Unknown argument provided!'

name_cannot_already_exist = '''
Invalid Argument: Cannot create a project in a directory that already exists!

    The directory "./{}" already exists. If you want to use this folder with
    mekpie, navigate within the folder, and run the `init` command.
'''

name_cannot_be_empty = '''
Missing argument: You must provide the positional argument `name`!

    When creating a project with `new` or `init` you must provide a positional
    argument `name` after the command.
'''

execution_error = '''
Execution error: Error while trying to execute a python resource!
    
    Error occured when attempting to execute: {}

{}
'''

compiler_config_error = '''
Config error: Invalid compiler configuration!

    Exptected one of the following: {}
    Instead found a value of: {}
'''

type_error = '''
Config error: Invalid type!

    Expected `{}` to be of type: {}
    Instead found a value of type: {}
{}
'''

file_not_found = '''
File not found: Could not find "{}"!
'''

directory_not_found = '''
Directory not found: Could not find "{}"!
'''

compiler_flag_error = '''
Flags not found: Could not find compiler flag for "{}" with purpose "{}"!

    Add the following entry to mek.py to include the correct flag:
    {}
'''

no_tests = '''
No tests found: Could not find any tests to run in /tests/!

    Try adding a c source file with a `main` function to /tests/
'''

no_tests_with_name = '''
No tests found: Could not find any test with the name {} in /tests/!

     Check that you spelled the test name correctly.
'''

api_no_options = '''
No options: API call to mekpie was made without an `Options` instance!

     Call `core.mekpie` with an instance of `definitions.Options`
'''

failed_compiler_call = '''
Build failed: the following call to the compiler failed:

    {}

And the following errors were produced:
    {}
'''

failed_program_call = '''
Call failed: the following program call failed:

    {}
'''

failed_autodetect = '''
Compiler detection failed: using a default configuration!

    Check mek.py and ensure `cmd` and and `dbg` are set to appropriate values
'''

release_debug = '''
Build failed: cannot launch debugger for a release build

    To launch the debugger run the command again without the --release flag.
'''

created = '''
{} created successfully!
'''

initialized = '''
{} initialized successfully!
'''

clean = '''
Project successfully cleaned.
'''

build_succeeded = '''
Project successfully built. ({:.3f}s)
'''


import mekpie.debug    as debug
import mekpie.messages as messages

from .create      import command_new, command_init
from .definitions import Options, Option

from .util import (
    car,
    cdr,
    tab,
    split,
    empty,
    panic,
    underline,
)
from .compiler import (
    command_run,
    command_test,
    command_clean,
    command_build,
    command_debug,
    command_dist,
)

def command_help(cfg):
    print(messages.usage)

def command_version(cfg):
    print(messages.version)

def pre_config_commands():
    return [
        command_help,
        command_version,
        command_new,
        command_init,
    ]

def default_options():
    return Options(
        quiet       = False,
        verbose     = False,
        release     = False,
        developer   = False,
        changedir   = False,
        command     = None,
        commandargs = [],
        programargs = [],
    )

def available_options():
    return [
        flag('quiet',            ['-q', '--quiet']),
        flag('verbose',          ['-v', '--verbose']),
        flag('release',          ['-r', '--release']),
        flag('developer',        ['-d', '--developer']),
        flag('changedir',        ['-c', '--changedir'], 2),
        command(command_help,    ['-h', '--help', 'help']),
        command(command_version, ['-V', '--version', 'version']),
        command(command_new,     ['new']),
        command(command_init,    ['init']),
        command(command_clean,   ['clean']),
        command(command_build,   ['build']),
        command(command_run,     ['run']),
        command(command_test,    ['test']),
        command(command_debug,   ['debug']),
        command(command_dist,    ['dist']),
    ]

def parse_arguments(args):
    argsall = args[:]
    options = default_options()._asdict()
    args, programargs = split(args, '--')
    options['programargs'] = programargs
    while not empty(args):
        arg = car(args)
        for names, nargs, handler in available_options():
            if arg in names:
                handler(options, args[:nargs], argsall)
                args = args[nargs:] if nargs else []
                break
        else:
            argument_error(messages.unknown_argument, car(args), argsall)
    return Options(**options)

def flag(name, aliases, nargs=1):

    def handle_flag(options, args, argsall):
        if options[name]:
            argument_error(messages.repeated_option.format(name), name, argsall)
        else:
            options[name] = cdr(args) or True

    return Option(
        names   = aliases,
        nargs   = nargs,
        handler = handle_flag,
    )

def command(command, aliases):

    def handle_command(options, args, argsall):
        if options['command']:
            argument_error(messages.too_many_arguments, command, argsall)
        else:
            options['commandargs'] = cdr(args)
            options['command']    = command

    return Option(
        names   = aliases,
        nargs   = None,
        handler = handle_command,
    )

def argument_error(message, arg, args):
    args = ['mekpie'] + args
    panic(f'{message}\n{tab(underline(arg, args))}')


from os.path import isdir, curdir, join
from os      import chdir
from os.path import basename, abspath, exists

import mekpie.messages as messages

from .util        import panic, empty, car, smkdir, exec_str
from .config      import config_from_dict
from .autodetect  import autodetect_compiler
from .definitions import DEFAULT_MEKPY, MAIN
from .structure   import (
    set_project_path,
    get_project_path,
    get_mekpy_path,
    get_src_path,
    get_main_path,
    get_test_path,
    get_includes_path,
    get_target_path,
    get_target_debug_path,
    get_target_release_path,
)

def command_new(options):
    name = project_name(options)
    check_name(name)
    create_project_directory(name)
    create_mekpy(name)
    create_src(name)
    create_tests()
    create_includes()
    create_target()
    config_from_dict(exec_str(get_mekpy_source(name), 'default mek.py', { 'options': options }), options)
    print(messages.created.format(name).strip())

def command_init(options):
    name = basename(abspath(curdir))
    check_name(name)
    create_mekpy(name)
    create_src(name)
    create_tests()
    create_includes()
    create_target()
    config_from_dict(exec_str(get_mekpy_source(name), 'default mek.py', { 'options': options }), options)
    print(messages.initialized.format(name).strip())

def check_name(name):
    if not name:
        panic(messages.name_cannot_be_empty)
    if isdir(name):
        panic(messages.name_cannot_already_exist.format(name))

def create_project_directory(name):
    set_project_path(name)
    smkdir(get_project_path())

def create_mekpy(name):
    if not exists(get_mekpy_path()):
        with open(get_mekpy_path(), 'w+') as rsc:
            rsc.write(get_mekpy_source(name))

def create_src(name):
    smkdir(get_src_path())
    if not exists(get_main_path(name + '.c')):
        with open(get_main_path(name + '.c'), 'w+') as rsc:
            rsc.write(get_main_source())

def create_tests():
    smkdir(get_test_path())

def create_includes():
    smkdir(get_includes_path())

def create_target():
    smkdir(get_target_path())
    smkdir(get_target_release_path())
    smkdir(get_target_debug_path())

def project_name(options):
    return car(options.commandargs)

def get_main_source():
    return MAIN + '\n'

def get_mekpy_source(name):
    compiler_preset = autodetect_compiler()
    return DEFAULT_MEKPY.format(
        name, 
        name + '.c', 
        compiler_preset.cc, 
        compiler_preset.cmd, 
        compiler_preset.dbg,
        compiler_preset.release_flags,
        compiler_preset.dbg_flags,
    )


from sys import version_info, stderr

# Require python 3.6
required_major = 3
required_minor = 6

if version_info.major < required_major or version_info.minor < required_minor:
    stderr.write('\nPython error: mekpie requires python version 3.6 or later!\n')
    exit(1)

from .core import main

# Program entry point for module
if __name__ == "__main__":
    main()


from os      import curdir, chdir, curdir
from sys     import argv
from os.path import abspath

import mekpie.debug    as debug
import mekpie.messages as messages

from .util        import log, car, cdr, last, panic, file_as_str, check_is_dir
from .config      import get_config
from .arguments   import parse_arguments, pre_config_commands
from .definitions import Options

def mekpie(options):
    root = abspath(curdir)
    if type(options) != Options:
        panic(messages.api_no_options)
    prepare_for_command(options)
    perform_command(options)
    chdir(root)

def main(args=cdr(argv)):
    debug.args = args
    mekpie(parse_arguments(args))

def prepare_for_command(options):
    if options.developer:
        debug.enable()
    if options.changedir:
        chdir(check_is_dir(car(options.changedir)))

def perform_command(options):
    command = options.command
    if command is None:
        panic(messages.no_command)
    elif command in pre_config_commands():
        command(log(options))
    else:
        config = get_config(options)
        command(log(config))


from os.path import join, curdir

project_path = curdir

def set_project_path(name):
    global project_path
    project_path = join(curdir, name)

def get_project_path():
    return project_path

def get_mekpy_path():
    return join(get_project_path(), 'mek.py')

def get_src_path():
    return join(get_project_path(), 'src')

def get_test_path():
    return join(get_project_path(), 'tests')

def get_includes_path():
    return join(get_project_path(), 'includes')

def get_target_path():
    return join(get_project_path(), 'target')

def get_target_debug_path():
    return join(get_target_path(), 'debug')

def get_target_release_path():
    return join(get_target_path(), 'release')

def get_main_path(main):
    return join(get_src_path(), main)
    


from os     import name
from shutil import which

import mekpie.messages as messages

from .util        import panic, log, errprint
from .definitions import CC_CMDS

warned = False

def autodetect_compiler():
    global warned
    for cmd in CC_CMDS.keys():
        if defined(cmd):
            return CC_CMDS[cmd]
    if not warned:
        errprint(messages.failed_autodetect + '\n')
        warned = True
    return CC_CMDS['default']

def defined(name):
    return which(name) is not None


from collections import namedtuple

Option = namedtuple('Option', [
    'names',
    'nargs',
    'handler',
])

Options = namedtuple('Options', [
    'quiet',       # -q --quiet
    'verbose',     # -v --verbose
    'release',     # -r --release
    'developer',   # -d --developer
    'changedir',   # -c --changedir
    'command',     # new, init, clean, build, run, test, debug
    'commandargs', # <optionargs...>
    'programargs', # -- <programargs...>
])

Config = namedtuple('Config', [
    'name',                   # Project name
    'main',                   # Entry point
    'libs',                   # Libraries to link
    'cc',                     # C Compiler Configuration
    'cmd',                    # The C Compiler command
    'dbg',                    # Debugger
    'flags',                  # User compiler flags
    'options',                # The provided command line options
])

CompilerPreset = namedtuple('CompilerPreset', [
    'cc',           # the compiler config
    'cmd',          # the compiler command
    'dbg',          # the debug command
    'dbg_flags',    # the debug flags
    'release_flags' # the release flags
])

CC_CMDS = {
    'clang' : CompilerPreset(
        cc            = 'gcc/clang',
        cmd           = 'clang',
        dbg           = 'lldb',
        dbg_flags     = '[\'-g\']',
        release_flags = '[\'-O\']',
    ),
    'gcc' : CompilerPreset(
        cc            = 'gcc/clang',
        cmd           = 'gcc',
        dbg           = 'gdb',
        dbg_flags     = '[\'-g\']',
        release_flags = '[\'-O\']',
    ),
    'default' : CompilerPreset(
        cc            = 'gcc/clang',
        cmd           = 'cc',
        dbg           = 'debugger',
        dbg_flags     = '[\'-g\']',
        release_flags = '[\'-O\']',
    )
}

DEFAULT_MEKPY='''
# This is a standard configuration file for mekpie

# the name of your project
name = '{}' 
# the .c file containing `main`
main = '{}'
# any libraries to load
libs = []
# the c compiler configuration to use (gcc/clang)
cc = '{}'
# the c compiler command to use on the command line
cmd = '{}'
# the debugger to use
dbg = '{}'
# additional compiler flags
flags = ['-Wall']

if options.release:
    flags = flags + {}
else:
    flags = flags + {}
'''

MAIN = '''
#include <stdio.h>
#include <stdlib.h>

int main() {
    puts("Hello, World!");
    return EXIT_SUCCESS;
}
'''.strip()

VERSION = '0.0.1'


from time               import time
from os.path            import join
from collections        import namedtuple
from concurrent.futures import ThreadPoolExecutor

import mekpie.messages as messages

from .util         import (
    smv,
    panic, 
    list_files, 
    list_all_dirs,
    remove_contents, 
    filename, 
)
from .runner       import lrun
from .structure    import (
    get_project_path,
    get_test_path,
    get_target_path,
    get_target_debug_path,
    get_target_release_path,
    get_main_path,
    get_src_path,
    get_includes_path,
)

max_threads = 32

def command_clean(cfg):
    remove_contents(get_target_debug_path())
    remove_contents(get_target_release_path())
    if not cfg.options.quiet:
        print(messages.clean.strip())

def command_build(cfg):
    command_clean(cfg)
    start   = time()
    sources = get_sources(cfg)
    mains   = [get_main_path(cfg.main), *list_files(get_test_path(), with_ext='.c')]
    compiler_configs[cfg.cc](cfg, sources, mains)
    if not cfg.options.quiet:
        print(messages.build_succeeded.format(time() - start).strip())

def command_run(cfg):
    command_build(cfg)
    lrun([get_bin_path(cfg, get_main_path(cfg.main))] + cfg.options.programargs, error=False)

def command_debug(cfg):
    if cfg.options.release:
        panic(messages.release_debug)
    command_build(cfg)
    lrun([cfg.dbg, get_bin_path(cfg, get_main_path(cfg.main))])

def command_test(cfg):
    command_build(cfg)
    for test in get_tests(cfg):
        lrun([get_bin_path(cfg, test)])

def get_tests(cfg):
    return list_files(get_test_path(), with_filter=lambda test : 
        test.endswith('.c')  and 
        (filename(test) in cfg.options.commandargs or len(cfg.options.commandargs) == 0)
    )

def command_dist(cfg):
    command_build(cfg)
    smv(
        get_bin_path(cfg, get_main_path(cfg.main)), 
        join(get_project_path(), cfg.name)
    )

def get_bin_path(cfg, path):
    root = get_target_release_path() if cfg.options.release else get_target_debug_path()
    return join(root, filename(path))

def get_sources(cfg):
    sources = list_files(get_src_path(), with_ext='.c')
    sources.remove(get_main_path(cfg.main))
    return sources

def get_includes_paths():
    includes = get_includes_path()
    return [includes] + list_all_dirs(includes)

# Compiler Configs
# ---------------------------------------------------------------------------- #

def gcc_clang_config(cfg, sources, mains):
    sources  = sources
    verbose  = ['-v'] if cfg.options.verbose else []
    libs     = ['-l' + lib for lib in cfg.libs]
    includes = ['-I' + inc for inc in get_includes_paths()]
    flags    = cfg.flags + verbose + includes
    objects  = [get_bin_path(cfg, source) + '.o'
        for source 
        in sources
    ]
    # Build objects
    with ThreadPoolExecutor(max_workers=max(1, min(len(sources), max_threads))) as e:
        results = e.map(lambda source: lrun([
            cfg.cmd, 
            *flags, 
            '-c', source,
            '-o', get_bin_path(cfg, source) + '.o'
        ], quiet=cfg.options.quiet, error=False), sources)

    if not all(results):
        exit(1)


    flags = flags + libs
    # Link executables
    with ThreadPoolExecutor(max_workers=max(1, min(len(mains), max_threads))) as e:
        results = e.map(lambda main : lrun([
            cfg.cmd,
            main,
            *objects,
            *flags, 
            '-o' , get_bin_path(cfg, main)
        ], quiet=True, error=False), mains)

    if not all(results):
        for main in mains:
            lrun([
                cfg.cmd,
                main,
                *objects,
                *flags,
                '-o' , get_bin_path(cfg, main)
            ], quiet=cfg.options.quiet)
        
compiler_configs = {
    'gcc/clang' : gcc_clang_config,
}